package com.fossgalaxy.games.tbs.ai.ga.macro.variable;

import com.fossgalaxy.games.tbs.GameState;
import com.fossgalaxy.games.tbs.ai.Controller;
import com.fossgalaxy.games.tbs.ai.rules.ProductionRule;
import com.fossgalaxy.games.tbs.ai.rules.ProductionRuleAgent;
import com.fossgalaxy.games.tbs.entity.Entity;
import com.fossgalaxy.games.tbs.order.Order;
import com.fossgalaxy.games.tbs.parameters.EntityType;
import com.fossgalaxy.object.annotations.ObjectDef;
import rts.ai.evaluation.EvaluationFunction;

import java.util.*;
import java.util.stream.Collectors;


/**
 * Solutions is a int[][] of Entity to indices to strategies
 */
public class VariableLengthMacroActionGA implements Controller {

    protected final Map<EntityType, List<ProductionRuleAgent>> rules;
    private final Map<EntityType, Integer> typeIndices;
    private final Controller opponent;
    private final Controller fallback;
    private final EvaluationFunction eval;

    private int horizon;
    private final int maxHorizon;

    private ActionSequence currentSequence;
    private ActionSequence nextSequence;
    private double nextScore;
    private int currentTick;
    private final int[] ruleLimits;
    private final int budget;

    @ObjectDef("VLMAGA")
    public VariableLengthMacroActionGA(int budget, int maxHorizon, EvaluationFunction eval, EntityType[] types, ProductionRuleAgent[][] strategies, Controller fallback, Controller opponent) {
        this.rules = new HashMap<>();
        for (int i = 0; i < types.length; i++) {
            Objects.requireNonNull(types[i]);
            this.rules.put(types[i], Arrays.asList(strategies[i]));
        }
        this.eval = eval;
        this.maxHorizon = maxHorizon;
        this.horizon = 1;
        this.fallback = fallback;
        this.opponent = opponent;
        this.budget = budget;

        typeIndices = new HashMap<>();
        ruleLimits = new int[rules.size()];
        int i = 0;
        for (EntityType type : rules.keySet()) {
            typeIndices.put(type, i);
            ruleLimits[i] = rules.get(type).size();
            i++;
        }
    }

    @Override
    public Map<UUID, Order> doTurn(int playerID, GameState state) {
        // Recompute!
        if (nextSequence == null) {
            nextSequence = new ActionSequence(rules.size(), ruleLimits, 5, 1, horizon);
        }
        long endTime = System.currentTimeMillis() + budget;
        GameState workingState = new GameState(state);
        if(currentSequence != null) {
            // Forward game state to the right place
            for (int i = currentTick; i < horizon; i++) {
                doSingleTurn(playerID, workingState, currentSequence, i);
            }
        }

        // Work on nextSequence
        int iterations = 0;
        while (System.currentTimeMillis() < endTime) {
            singleIteration(new GameState(workingState), playerID);
            iterations++;
        }
        System.out.println("Iterations achieved: " + iterations);

        if(currentSequence == null){
            currentSequence = nextSequence;
        }
        Map<UUID, Order> orders = doSingleTurn(playerID, state, currentSequence, currentTick++);

        // We have reached the end of the currentSequence
        if (currentTick == horizon) {
            currentTick = 0;
            currentSequence = nextSequence;
            System.out.println("Phase over");
            System.out.println(currentSequence.print(rules, typeIndices));
            if (horizon < maxHorizon) {
                horizon = Math.min(maxHorizon, horizon + 2);
                System.out.println("New Horizon: " + horizon);
            }
            nextSequence = null;
        }

        return orders;
    }

    private void singleIteration(GameState state, int playerID) {
        ActionSequence child = new ActionSequence(nextSequence);
        child.mutate(0.1, 0.1);
        simulate(state, child, playerID);
        double childScore = eval.evaluate(playerID, getNextPlayer(state, playerID), state);
        if (childScore > nextScore) {
            nextSequence = child;
            nextScore = childScore;
        }
    }

    private int getNextPlayer(GameState state, int playerID) {
        return (playerID + 1) % state.getNumberOfPlayers();
    }

    private GameState simulate(GameState state, ActionSequence actionSequence, int playerID) {
        for (int actionIndex = 0; actionIndex < actionSequence.length() && !state.isGameOver(); actionIndex++) {
            doSingleTurn(playerID, state, actionSequence, actionIndex);
        }
        return state;
    }

    private Map<UUID, Order> doSingleTurn(int playerID, GameState state, ActionSequence actionSequence, int actionIndex) {
        Collection<Entity> entities = state.getOwnedEntities(playerID);
        Map<UUID, Order> ordersSoFar = doAutomatic(entities, state);
        for (Map.Entry<EntityType, List<ProductionRuleAgent>> entry : rules.entrySet()) {
            EntityType type = entry.getKey();
            int typeIndex = typeIndices.get(type);
            int strategy = actionSequence.getStrategy(typeIndex, actionIndex);
            ProductionRuleAgent controller = rules.get(type).get(strategy);
            Map<UUID, Order> ordersThisType = controller.doTurn(playerID, state);
            for (Map.Entry<UUID, Order> orderEntry : ordersThisType.entrySet()) {
                UUID id = orderEntry.getKey();
                Order order = orderEntry.getValue();
                if (!ordersSoFar.containsKey(id) && state.getEntityByID(id).getType().isInstance(type)) {
                    ordersSoFar.put(id, order);
                    order.doOrder(state.getEntityByID(id), state);
                }
            }
        }

        // Fallback time
        Map<UUID, Order> fallbackOrders = fallback.doTurn(playerID, state);
        for(Map.Entry<UUID, Order> entry : fallbackOrders.entrySet()){
            UUID id = entry.getKey();
            Order order = entry.getValue();
            ordersSoFar.putIfAbsent(entry.getKey(), order);
            order.doOrder(state.getEntityByID(id), state);
        }

        // Simulate opponent
        Map<UUID, Order> opponentOrders = opponent.doTurn(getNextPlayer(state, playerID), state);
        for (Map.Entry<UUID, Order> entry : opponentOrders.entrySet()) {
            entry.getValue().doOrder(state.getEntityByID(entry.getKey()), state);
        }

        return ordersSoFar;
    }


    /**
     * Scan and find entities that can have automatic actions applied without bothering the GA.
     *
     * @param entitiesNeedingOrders The entities in the game that are ours
     * @return Orders generated from this
     */
    protected Map<UUID, Order> doAutomatic(Collection<Entity> entitiesNeedingOrders, GameState state) {
        return entitiesNeedingOrders.stream()
                .filter(x -> x.getType().getAvailableActions().size() == 1)
                .filter(x -> x.getType().getAvailableActions().get(0).canAutomate())
                .collect(Collectors.toMap(Entity::getID, x -> x.getType().getAvailableActions().get(0).generateOrder(x.getPos(), state)));
    }
}

class ActionSequence {
    // Great for one entity
    //lengths[x][y] refers to x EntityType, y = length of strategy
    //strategies[x][y] refers to x EntityType, y = indexOfStrategy
    private int[][] lengths;
    private int[][] strategies;

    // Min and max values for lengths array
    private final int maxMacroLength;
    private final int minMacroLength;
    private final int[] ruleLimits;
    private final Random random;
    private final int numEntityTypes;
    private final int numberOfActions;

    public ActionSequence(int numEntityTypes, int[] ruleLimits, int maxMacroLength, int minMacroLength, int numberOfActions) {
        this.maxMacroLength = maxMacroLength;
        this.minMacroLength = minMacroLength;
        this.ruleLimits = ruleLimits;
        this.numEntityTypes = numEntityTypes;
        this.random = new Random();
        this.numberOfActions = numberOfActions;
        this.lengths = new int[numEntityTypes][numberOfActions];
        this.strategies = new int[numEntityTypes][numberOfActions];
        randomize();
    }

    public ActionSequence(ActionSequence parent) {
        this.maxMacroLength = parent.maxMacroLength;
        this.minMacroLength = parent.minMacroLength;
        this.ruleLimits = Arrays.copyOf(parent.ruleLimits, parent.ruleLimits.length);
        this.random = new Random();
        this.numEntityTypes = parent.numEntityTypes;
        this.numberOfActions = parent.numberOfActions;
        this.lengths = new int[numEntityTypes][numberOfActions];
        this.strategies = new int[numEntityTypes][numberOfActions];
        for (int i = 0; i < lengths.length; i++) {
            this.lengths[i] = Arrays.copyOf(parent.lengths[i], numberOfActions);
            this.strategies[i] = Arrays.copyOf(parent.strategies[i], numberOfActions);
        }
    }

    private void randomize() {
        for (int entityIndex = 0; entityIndex < numEntityTypes; entityIndex++) {
            for (int actionIndex = 0; actionIndex < lengths[entityIndex].length; actionIndex++) {
                lengths[entityIndex][actionIndex] = randomLength();
                strategies[entityIndex][actionIndex] = randomStrategyIndex(entityIndex);
            }
        }
    }

    private int randomLength() {
        return random.nextInt(maxMacroLength - minMacroLength) + minMacroLength;
    }

    private int randomStrategyIndex(int entityIndex) {
        return random.nextInt(ruleLimits[entityIndex]);
    }

    public void mutate(double lengthChance, double strategyChance) {
        for (int entityIndex = 0; entityIndex < numEntityTypes; entityIndex++) {
            // Lengths
            for (int lengthIndex = 0; lengthIndex < lengths[entityIndex].length; lengthIndex++) {
                if (random.nextDouble() < lengthChance) {
                    lengths[entityIndex][lengthIndex] = randomLength();
                }
            }
            // Strategies
            for (int strategyIndex = 0; strategyIndex < strategies[entityIndex].length; strategyIndex++) {
                if (random.nextDouble() < strategyChance) {
                    strategies[entityIndex][strategyIndex] = randomStrategyIndex(entityIndex);
                }
            }
        }
    }

    public int length() {
        return lengths.length;
    }

    public int getStrategy(int entityIndex, int actionIndex) {
        int runningTotal = 0;
        for (int i = 0; i < lengths[entityIndex].length; i++) {
            runningTotal += lengths[entityIndex][i];
            if (actionIndex < runningTotal) return strategies[entityIndex][i];
        }
        return 0;
    }

    public String print(Map<EntityType, List<ProductionRuleAgent>> rules, Map<EntityType, Integer> typeIndices){
        StringBuilder builder = new StringBuilder();
        for(Map.Entry<EntityType, Integer> entry : typeIndices.entrySet()){
            EntityType type = entry.getKey();
            int entityIndex = entry.getValue();

            builder.append(type.getName());
            builder.append(": ");
            int[] strategy = strategies[entityIndex];
            int[] individualLengths = lengths[entityIndex];
            for(int strategyIndex = 0; strategyIndex < strategy.length; strategyIndex++){
                if(strategyIndex != 0) builder.append(",");
                builder.append(rules.get(type).get(strategy[strategyIndex]).getName());
                builder.append(" for ");
                builder.append(individualLengths[0]);
                builder.append(" turns");
            }
            builder.append("\n");
        }
        return builder.toString();
    }
}
