package com.fossgalaxy.games.tbs.order;

import com.fossgalaxy.games.tbs.GameState;
import com.fossgalaxy.games.tbs.actions.BuildAction;
import com.fossgalaxy.games.tbs.entity.Entity;
import com.fossgalaxy.games.tbs.parameters.EntityType;
import com.fossgalaxy.games.tbs.parameters.TerrainType;
import com.fossgalaxy.games.tbs.ui.GameAction;
import org.codetome.hexameter.core.api.CubeCoordinate;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Random;

public class BuildOrder implements Order {
    private EntityType type;
    private CubeCoordinate cube;
    private Random random;

    public BuildOrder(EntityType type, CubeCoordinate cube) {
        this.type = type;
        this.cube = cube;
        this.random = new Random();
    }

    @Override
    public void doOrder(Entity host, GameState state) {
        if (!substituteType(host)) return;
        if (!isNextToMe(host, state)) return;
        if (!isTerrainPassableForBuildTypeOrShouldTheyGoHome(state)) return;
        if (isSpaceEmpty(state)) return;
        if (!payIfPossible(host, state)) return;
        build(host, state);
    }

    /**
     * Might not be able to build this type exactly - do we have another BuildOrder that can?
     * Will find either that the correct type is buildable by this unit,
     * or select a random sub-type of the type given
     *
     * @param host The host entity doing the production
     */
    protected boolean substituteType(Entity host) {
        List<GameAction> actions = host.getType().getAvailableActions();
        List<EntityType> viableTypes = new ArrayList<>();
        for (GameAction action : actions) {
            if (action instanceof BuildAction) {
                BuildAction build = (BuildAction) action;
                EntityType buildType = build.getProductionType();
                if (buildType.equals(type)) return true;
                if (buildType.isInstance(type)) {
                    viableTypes.add(buildType);
                }
            }
        }
        if (viableTypes.isEmpty()) return false;
        type = viableTypes.get(random.nextInt(viableTypes.size()));
        return true;
    }

    protected boolean isNextToMe(Entity host, GameState state) {
        return state.getDistance(host.getPos(), cube) == BuildAction.BUILD_RANGE;
    }

    protected boolean isTerrainPassableForBuildTypeOrShouldTheyGoHome(GameState state) {
        TerrainType tt = state.getTerrainAt(cube);
        if (tt == null) return false;
        return state.getTerrainAt(cube).isPassible(type);
    }

    protected void build(Entity host, GameState state) {
        // build it
        Entity e = new Entity(type, cube, host.getOwner());
        state.addEntity(e);
    }

    protected boolean payIfPossible(Entity host, GameState state) {
        //check we can afford to by the object
        Map<String, Integer> cost = type.getCosts();
        for (Map.Entry<String, Integer> costEntry : cost.entrySet()) {
            String t = costEntry.getKey();
            int costVal = costEntry.getValue();

            int currStore = state.getResource(host.getOwner(), t);
            if (currStore < costVal) {
                return false;
            }
        }

        // pay for it
        for (Map.Entry<String, Integer> costEntry : cost.entrySet()) {
            String t = costEntry.getKey();
            int costVal = costEntry.getValue();
            state.addResource(host.getOwner(), t, -costVal);
        }
        return true;
    }

    protected boolean isSpaceEmpty(GameState state) {
        // check the space we're building in is free
        Entity entity = state.getEntityAt(cube);
        if (entity != null) {
            return true;
        }
        return false;
    }

    public String toString() {
        return String.format("build %s at (%d, %d)", type, cube.getGridX(), cube.getGridZ());
    }

}
